local shortName,N = ...
----------------
-- File globals
----------------
local select, LibStub
    = select, LibStub

local A = LibStub("Abacus-1.0")
local D = LibStub("DoomCore-2.0")
local updateFrame, setFont = D.updateFrame, D.setFont
local setFrameEnabled = A.setFrameEnabled
N.updateFrame = updateFrame
local Addon = D.Addon(shortName, "Conversation", N)
Addon.version = 0.2
local setIndex = A.setIndex

local lsm = LibStub("LibSharedMedia-3.0")

local frames = {
  conversation = Conversation,
  chat = ConversationChat
}
--------
-- Core
--------

local managed = { "QuestFrame", "GossipFrame", "QuestInfoFrame", "QuestInfoRewardsFrame" }
function N.Show()
  UIParent:Show()
  for _,manage in pairs(managed) do setFrameEnabled(manage, true) end
end
function N.Hide()
  for _,manage in pairs(managed) do setFrameEnabled(manage, false) end
end
DEBSHOW = N.Show
DEBHIDE = N.Hide

function Addon:OnInitialize()
  self:Register(nil, "Default")
  lsm:RegisterCallback("LibSharedMedia_Registered", function()
    self:Rebuild()
  end)
  self:TrackEvent("TALKINGHEAD_REQUESTED")
  N.Hide()
end

function Addon:OnLoad()
  local core, defaults = self.core, self.defaults
  self:Inherit("talkinghead", "conversation")
  self:Inherit("chat", "conversation")
  if not core.conversation then
    self:Make(core, "conversation", "conversation")
    local conversation = core.conversation
    for k,v in pairs(core) do
      if type(v) ~= "table" or not v.type then conversation[k] = v end
    end
    for k,v in pairs(core) do
      if type(v) ~= "table" or not v.type then core[k] = nil end
    end
  end
  if not core.talkinghead then self:Make(core,"talkinghead", "talkinghead") end
  setIndex(defaults.talkinghead, core.conversation)
  if not core.chat then self:Make(core,"chat", "chat") end
  setIndex(defaults.chat, core.conversation)
  self:Rebuild()
end


------------------
-- Frames
------------------

function Addon:BuildChoice(choice, conf)
  conf = conf or self.core.conversation
  setFont(choice.TextFrame.Text, {
    font = conf.choiceFont,
    fontSize = conf.choiceSize
  })
  setFont(choice.TextFrame.TextVirtual, {
    font = conf.choiceFont,
    fontSize = conf.choiceSize
  })
end

local rebuildHead

function Addon:Rebuild(info, r, g, b, a)
  if info then self:ConfSet(info, r, g, b, a) end
  rebuildHead = true
  local core = self.core
  local conversationConf = core.conversation
  setFrameEnabled("ItemTextFrame", not conversationConf.read)

  for _,part in pairs{ "conversation", "chat" } do
    local frame = frames[part]
    local conf = core[part]
    frame.TextFrame.Progress:GetAnimations():SetDuration(conf.slideSpeed)
    updateFrame(frame,conf)
    frame:AnchorAlerts(conf)
    setFont(frame.NameFrame.Name, {
      font = conf.titleFont,
      fontSize = conf.titleSize
    })
    setFont(frame.TextFrame.Text, {
      font = conf.textFont,
      fontSize = conf.textSize
    })
  end
  local choices = frames.conversation.ChoicesFrame
  for i=1,select("#", choices:GetChildren()) do
    self:BuildChoice(select(i, choices:GetChildren()), conversationConf)
  end
  frames.conversation:UpdateFriendship()
end

function Addon:TALKINGHEAD_REQUESTED()
  local head = TalkingHeadFrame
  if not head or not rebuildHead then return end
  rebuildHead = nil
  local conf = self.core.talkinghead
  head:SetScale(conf.scale)
  head:SetWidth(conf.width)
  setFont(head.NameFrame.Name, {
    font = conf.titleFont,
    fontSize = conf.titleSize
  })
  setFont(head.TextFrame.Text, {
    font = conf.textFont,
    fontSize = conf.textSize
  })
end

function Addon:Close()
  self:Output("Closing")
  frame:Close()
end