local A = LibStub:NewLibrary("Abacus-1.0",3)
if not A then return end
tconcat = tconcat or table.concat
local tconcat = tconcat
pow = pow or math.pow
local pow = pow
fmod = fmod or math.fmod
local fmod = fmod

local ceil, floor, getn, ipairs, min, max, next, pairs, select, strsplit, tonumber, tostring, type, unpack
= ceil, floor, getn, ipairs, min, max, next, pairs, select, strsplit, tonumber, tostring, type, unpack
local error, getmetatable, print, setmetatable, sort, tinsert
= error, getmetatable, print, setmetatable, sort, tinsert
local GameTooltip, GetItemInfo, GetSpellInfo
= GameTooltip, GetItemInfo, GetSpellInfo


local function tcontains(list,el)
  for k,v in pairs(list) do
    if v == el then return k,v end
  end
end
A.tcontains = tcontains

local function bounded(minBound, val, maxBound)
  if minBound and val < minBound then return minBound
  elseif maxBound and val > maxBound then return maxBound
  else return val
  end
end
A.bounded = bounded

local function isEmpty(val)
  return val == nil or val == "" or (type(val) == "table" and next(val) == nil)
end
A.isEmpty = isEmpty

local function tostrings(...)
  local output = {}
  for i=1,select("#", ...) do
    output[i] = tostring(select(i, ...))
  end
  return tconcat(output,"")
end
A.tostrings = tostrings

local function tovstring(val)
  if type(val) == "table" then
    if isEmpty(val) then return "{}"
    elseif #val > 0 then
      local output = {}
      for i=1,#val do
        output[i] = tostring(val[i])
      end
      return "{" .. tconcat(output, ", ") .. "}"
    else
      local output = {}
      for k,v in pairs(val) do
        tinsert(output, tostring(k) .. " = " .. tostring(v))
      end
      return "{" .. tconcat(output, ", ") .. "}"
    end
  end
  return tostring(val)
end
A.tovstring = tovstring

local function tovstrings(...)
  local output = {}
  for i=1,select("#", ...) do
    output[i] = tovstring(select(i, ...))
  end
  return tconcat(output,"")
end
A.tovstrings = tovstrings

local function tprint(table) for k,v in pairs(table) do print(k,v) end end
A.tprint = tprint

local function tstartswith(from, ...)
  for i=1,select("#", ...) do
    if from[i] ~= select(i, ...) then return false end
  end
  return true
end
A.tstartswith = tstartswith

local function tappend(t, ...)
  local len = #t
  for i=1,select("#", ...) do
    t[len + i] = select(i, ...)
  end
  return t
end
A.tappend = tappend

local function sublist(list, first, last)
  if first < 0 then first = 1 + #list + first end
  if last < 0 then last = 1 + #list + last end
  local toreturn = {}
  for i=0,last-first do toreturn[i + 1] = list[i + first] end
  return toreturn
end
A.sublist = sublist

local function assertNotNil(...)
  local failures = {}
  for i=1,select("#", ...) do
    if select(i, ...) == nil then
      tinsert(failures, tostrings("Arg #",ceil(i/2)," is nil."))
    end
  end
  if #failures>0 then error(tconcat(failures," "),2) end
  return ...
end
A.assertNotNil = assertNotNil

local function assertType(...)
  local failures = {}
  for i=1,select("#", ...) - 1,2 do
    local isType = type(select(i, ...))
    local matchType = select(i+1, ...)
    if not tcontains({ strsplit("/", matchType) }, isType) then
      tinsert(failures, tostrings("Arg #",ceil(i/2)," must be ",matchType,", is ",isType, "."))
    end
  end
  if #failures>0 then error(tconcat(failures," "), tonumber(select(-1, ...)) or 2) end
end
A.assertType = assertType

local function nilSort(val, valFunc, sortFunc)
  local reverse = false
  local valFuncType = type(valFunc)
  if valFuncType == "string" and valFunc:sub(1,1) == "-" then
    reverse = true
    if #valFunc < 2 then valFunc = nil else valFunc = valFunc:sub(2) end
  end
  sort(val, function(a, b)
    if valFuncType == "string" then
      if a ~= nil then a = a[valFunc] end
      if b ~= nil then b = b[valFunc] end
    elseif valFuncType == "function" then
      if a ~= nil then a = valFunc(a) end
      if b ~= nil then b = valFunc(b) end
    end
    if a == nil and b == nil then
      return false
    end
    if a == nil then
      return false
    end
    if b == nil then
      return true
    end
    if sortFunc then return sortFunc(a,b)
    elseif reverse then return b < a
    else return a < b
    end
  end)
end
A.nilSort = nilSort

local function capitalize(val)
  return val:sub(1, 1):upper() .. val:sub(2):lower()
end
A.capitalize = capitalize

local function abbrev(val, places)
  places = places or (val >= 1000000 and 2) or 0
  local place = pow(10, places)
  val = val * place
  if val >= 1000000 then
    return floor(val / 1000000) / place .. "m"
  elseif val >= 1000 then
    return floor(val / 1000) / place .. "k"
  else
    return floor(val)
  end
end
A.abbrev = abbrev

local function reverse(...)
  local l, t
  if select("#", ...) > 1 then
    l = { ... }
    t = true
  else
    l = ...
  end
  local newL = {}
  for i = getn(l), 1, -1 do tinsert(newL, l[i]) end
  if t then return unpack(newL) else return newL end
end
A.reverse = reverse

local function flip(flip, ...)
  if flip then return reverse(...) else return ... end
end
A.flip = flip

local function tupleify(func)
  return function(...)
    local len = select("#", ...)
    if len <= 1 then return func(...) end
    local r = {}
    for i=1,len do
      tinsert(r, func(...))
    end
    return unpack(r)
  end
end
A.tupleify = tupleify
local function scopify(func, ...)
  local scope
  local scopeCount = select("#", ...)
  if scopeCount > 0 then scope = { ... } end
  return function(...)
    if scope == nil then return func(...) end
    local count = select("#", ...)
    if count == 0 then return func(unpack(scope)) end
    local copy = { unpack(scope) }
    for i=1,count do
      copy[scopeCount + i - 1] = select(i, ...)
    end
    func(unpack(copy))
  end
end
A.scopify= scopify

local function map(func, before, subArgs, after)
  assertType(func,"function", before,"table/nil", subArgs,"table/nil", after,"table/nil", 3)
  local r = {}
  local result
  local args = { unpack(before or {}) }
  local slot = #args + 1
  if after then for i=1,#after do args[slot + i] = after[i] end end
  for i=1,#subArgs do
    local arg = subArgs[i]
    args[slot] = arg
    result = { func(unpack(args)) }
    if isEmpty(result) then result = nil elseif #result == 1 then result = result[1] end
    r[i] = result
  end
  return unpack(r)
end
A.map = map

local function colUnpack(col)
  if #col >= 3 then return unpack(col) end
  return col.r, col.g, col.b, col.a
end
A.colUnpack = colUnpack
local function colPack(r,g,b,a)
  return {r=r,g=g,b=b,a=a}
end
A.colPack = colPack

local function hexToCol(code)
  if code:sub(1,1) == "#" then code = code:sub(2) end
  if isEmpty(code) then return { r = 1, g = 1, b = 1, a = 1 } end
  return {
    r = tonumber("0x"..code:sub(1,2)) / 255,
    g = tonumber("0x"..code:sub(3,4)) / 255,
    b = tonumber("0x"..code:sub(5,6)) / 255,
    a = 1
  }
end
A.hexToCol = hexToCol
local function numToHex(num)
  local hexstr = "0123456789abcdef"
  local s = ""
  while num > 0 do
    local mod = fmod(num, 16)
    s = hexstr:sub(mod+1, mod+1) .. s
    num = floor(num / 16)
  end
  if s == "" then s = "0" end
  return s
end
A.numToHex = numToHex
local function colToHex(col)
  local r,g,b,a = colUnpack(col)
  local s = ""
  local lst = a and { a, r, g, b } or { r, g, b}
  for _,val in ipairs(lst) do
    local val = numToHex(val * 255)
    if #val < 2 then val = "0" .. val end
    s = s .. val
  end
  return s
end
A.colToHex = colToHex

local function colToHue(col)
  local r, g, b = colUnpack(col)
  if r == g and g == b then return 360 end
  local hue
  local minCol = min(r, g, b)
  local maxCol = max(r, g, b)
  local range = maxCol - minCol
  if maxCol == r then
    hue = (g - b) / range
  elseif maxCol == g then
    hue = 2 + (b - r) / range
  else
    hue = 4 + (r - g) / range
  end
  hue = hue * 60
  if hue < 0 then hue = hue + 360 end
  return hue
end
A.colToHue = colToHue
local function colorify(text,col)
  if type(col) ~= "string" then col = colToHex(col)
  elseif col:sub(1,1) == "#" then col = col:sub(2)
  end
  if #col == 6 then col = "FF" .. col end
  return "|c" .. col .. tostring(text) .. "|r"
end
A.colorify = colorify

local getItemID = tupleify(function(val)
  local itemString = select(2, GetItemInfo(val))
  if type(itemString) == "string" then
    local subString = itemString:sub(itemString:find(":") + 1)
    subString = subString:sub(1, subString:find(":") - 1)
    return tonumber(subString)
  end
end)
A.getItemID = getItemID

local function keybindmods()
  local text = ""
  if IsShiftKeyDown() then text = "SHIFT-" .. text end
  if IsControlKeyDown() then text = "CTRL-" .. text end
  if IsAltKeyDown() then text = "ALT-" .. text end
  return text
end
A.keybindmods = keybindmods

----------
-- Frames
----------

local function fitToText(button,dW,dH)
  local width,height = button:GetFontString():GetSize()
  button:SetSize(width + (dW or 0), height + (dH or 0))
end
A.fitToText = fitToText

local enable = tupleify(function(frame) frame:EnableMouse(true) end)
A.enable = enable

local function tooltip(button, show, anchor, onHide)
  button:SetMotionScriptsWhileDisabled(show)
  button:SetScript("OnEnter", function(button)
    if not button:IsShown() then return end
    if anchor then
      GameTooltip:SetOwner(button, "ANCHOR_" .. anchor)
    else
      GameTooltip:SetOwner(button)
    end
    local buttonType = button:GetAttribute("type")
    local subject = buttonType and button:GetAttribute(buttonType)
    if button.tooltip then
      GameTooltip:SetText(button.tooltip)
    elseif buttonType == "spell" then
      if type(subject) ~= "number" then
        subject = select(7, GetSpellInfo(subject))
      end
      GameTooltip:SetSpellByID(subject)
    elseif buttonType == "item" then
      GameTooltip:SetItemByID(subject)
    end
    GameTooltip:Show()
  end)
  button:SetScript("OnLeave", function(button)
    if not button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
  end)
  button:SetScript("OnHide", function(button)
    if button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
    if onHide then onHide(button) end
  end)
end
A.tooltip = tooltip

-------------------
-- Object-oriented
-------------------

local function setIndex(obj, ind, noOverride)
  local meta = getmetatable(obj)
  if meta == nil then
    meta = {}
    setmetatable(obj, meta)
  end
  if noOverride and meta.__index then return end
  meta.__index = ind
end
A.setIndex = setIndex

local Inherits = {}
function Inherits:New()
  self.__index = self
  local new = {}
  setmetatable(new, self)
  new._super = self
  return new
end
A.Inherits = Inherits

function Inherits:Map(func, before, subArgs, after)
  if type(func) ~= "function" then func = self[func] end
  before = { self, unpack(before or {}) }
  map(func, before, subArgs, after)
end

local function mixin(obj, ...)
  for i=1,select("#", ...) do
    for k,v in pairs(select(i, ...) or {}) do
      obj[k] = obj[k] or v
    end
  end
end
A.mixin = mixin
Inherits.MixIn = mixin

------------
-- Database
------------

local function npairs(val, depth, ...)
  if val == nil or type(val) ~= "table" then return pairs{} end
  local i = {}
  local done = false

  if depth <= 1 then return function()
    if done then return end
    done = true
    return val
  end end

  local subPairs = npairs(val, depth - 1, select(2, ...))
  local result = {}
  local check = ...

  local function assignNext()
    local subVal = result[depth - 1]
    if subVal == nil or type(subVal) ~= "table" then
      i[depth -1], i[depth] = nil, nil
      return
    end
    while true do
      i[depth -1], i[depth] = next(subVal, i[depth -1])
      if i[depth -1]==nil then return end
      local checkType = type(check)
      if not check and i[depth -1]~=nil then return end
      if checkType == "function" and check(i[depth]) then return end
    end
  end

  return function()
    assignNext()
    while not done and i[depth] == nil do
      result = { subPairs() }
      if result[1] == nil then
        done = true
        return
      else
        i = sublist(result, 1, -2)
        assignNext()
      end
    end
    if not done then return unpack(i) end
  end
end
A.npairs = npairs

-----------------
-- Bootstrapping
-----------------
local function buildfunc(before, oldfunc, after)
  if type(oldfunc) == "function" then
    if type(before) == "function" then
      if type(after) == "function" then
        return function(...) before(...); oldfunc(...); after(...) end
      else
        return function(...) before(...); oldfunc(...) end
      end
    else
      if type(after) == "function" then
        return function(...)              oldfunc(...); after(...) end
      end
    end
  else
    if type(before) == "function" then
      if type(after) == "function" then
        return function(...) before(...);               after(...) end
      else
        return function(...) before(...) end
      end
    else
      if type(after) == "function" then
        return function(...)                            after(...) end
      end
    end
  end
end
A.buildfunc = buildfunc

local function overfunc(parent, name, before, after)
  parent[name] = buildfunc(before, parent[name], after)
end
A.overfunc = overfunc

local function overscript(obj, script, before, after)
  obj:SetScript(script, overfunc(before, obj:GetScript(script), after))
end
A.overscript = overscript

local Hidden = CreateFrame("FRAME")
Hidden:Hide()

local frameanchors = {}
local framedata = {}
local frameparents = {}
local layouted = {}

local function disableFrame(frameName)
  local frame = _G[frameName]
  local parent = frame:GetParent()
  if parent == Hidden then return end
  frameanchors[frameName] = {}
  local anchors = frameanchors[frameName]
  for i=1,frame:GetNumPoints() do
    anchors[i] = { frame:GetPoint(i) }
  end
  layouted[frameName] = layouted[frameName] or frame:GetAttribute("UIPanelLayout-defined")
  frameparents[frameName] = parent
  --frame:SetAttribute("UIPanelLayout-defined", false)
  framedata[frameName] = UIPanelWindows[frameName] or framedata[frameName]
  --UIPanelWindows[frameName] = nil
  frame:SetParent(Hidden)
end
A.disableFrame = disableFrame

local function enableFrame(frameName)
  local frame = _G[frameName]
  if frame:GetParent() ~= Hidden then return end
  --frame:SetAttribute("UIPanelLayout-defined", layouted[frameName])
  --UIPanelWindows[frameName] = framedata[frameName] or UIPanelWindows[frameName]
  frame:SetParent(frameparents[frameName])
  --local anchors = frameanchors[frameName]
  --for i=1,#anchors do
  --frame:SetPoint(unpack(anchors[i]))
  --end
end
A.enableFrame = enableFrame

local function setFrameEnabled(frameName, display)
  if display then enableFrame(frameName) else disableFrame(frameName) end
end
A.setFrameEnabled = setFrameEnabled